﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Animation;

namespace UltimateCorp.Palette
{
    public abstract class BasePaletteAnimation : Control, INotifyPropertyChanged
    {
        protected const double MINIMUM_SIZE = 25.0D;

        protected bool DisableAnim = false;
        protected bool IsInitialized { get; set; } = false;

        [Flags]
        protected enum SetMode { Back = 0x01, Top = 0x02, All = Back | Top }

        #region events
        public event EndAnimationEventHandler EndAnim;
        public event PropertyChangedEventHandler? PropertyChanged;
        protected void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        private void OnEndAnim() => EndAnim?.Invoke();

        #region EndAnimation (ROUTED EVENT)
        public event RoutedEventHandler EndAnimation
        {
            add => AddHandler(PaletteContentControl.EndAnimationEvent, value);
            remove => RemoveHandler(PaletteContentControl.EndAnimationEvent, value);
        }
        public static readonly RoutedEvent EndAnimationEvent = EventManager.RegisterRoutedEvent(
            "EndAnimation",
            RoutingStrategy.Bubble,
            typeof(RoutedEventHandler),
            typeof(BasePaletteAnimation));
        protected void OnEndAnimation() => this.RaiseEvent(new RoutedEventArgs(PaletteTextControl.EndAnimationEvent));
        #endregion

        #endregion

        static BasePaletteAnimation()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(BasePaletteAnimation),
                new FrameworkPropertyMetadata(typeof(BasePaletteAnimation)));
        }

        #region ApplyTemplate
        protected BasePaletteControl paletteBack = null;
        protected BasePaletteControl paletteTop = null;
        protected BasePaletteControl paletteBottom = null;
        protected Border frontBottom = null;
        protected Border frontTop = null;
        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();
            paletteBack = GetTemplateChild("PART_Back") as BasePaletteControl;
            paletteTop = GetTemplateChild("PaletteTop") as BasePaletteControl;
            paletteBottom = GetTemplateChild("PaletteBottom") as BasePaletteControl;
            frontBottom = GetTemplateChild("PART_FrontBottom") as Border;
            frontTop = GetTemplateChild("PART_FrontTop") as Border;
            SetAll();
            paletteBack.SizeChanged += SizeChanged;
            this.Loaded += delegate
            {
                this.Dispatcher.BeginInvoke(new Action(() =>
                {
                    IsInitialized = true;
                }));
            };
        }
        #endregion

        protected void SizeChanged(object sender = null, SizeChangedEventArgs e = null)
        {
            double w = 0, h = 0;
            if (paletteBack != null &&
                paletteBottom != null &&
                paletteTop != null)
            {

                w = (paletteBack?.ActualWidth ?? 0D);
                h = (paletteBack?.ActualHeight ?? 0D) + AdjustTop;
            }
            if (w <= 0) w = MINIMUM_SIZE;
            if (h < 0) h = 0;
            else if (h == 0) h = MINIMUM_SIZE;
            this.MinHeight = h;
            this.MinWidth = w;
        }

        protected abstract UIElement GetContent { get; }
        protected abstract bool HasContent { get; }

        protected virtual void SetAll(SetMode mode = SetMode.All)
        {
            if (HasContent)
            {
                if ((mode & SetMode.Back) != 0)
                {
                    if (paletteBack != null) paletteBack.Content = GetContent;
                }
                if ((mode & SetMode.Top) != 0)
                {
                    if (paletteTop != null) paletteTop.Content = GetContent;
                    if (paletteBottom != null) paletteBottom.Content = GetContent;
                }
            }
            SizeChanged();
        }

        #region DP

        #region ShowLine (DP SHORT)
        [Category("Palette")]
        public bool ShowLine { get { return (bool)GetValue(ShowLineProperty); } set { SetValue(ShowLineProperty, value); } }
        public static readonly DependencyProperty ShowLineProperty = DependencyProperty.Register("ShowLine", typeof(bool), typeof(BasePaletteAnimation), new PropertyMetadata(false));
        #endregion

        #region LineColor (DP SHORT)
        [Category("Palette")]
        public Brush LineColor { get { return (Brush)GetValue(LineColorProperty); } set { SetValue(LineColorProperty, value); } }
        public static readonly DependencyProperty LineColorProperty = DependencyProperty.Register("LineColor", typeof(Brush), typeof(BasePaletteAnimation), new PropertyMetadata(Brushes.White));
        #endregion

        #region ContentLine (DP SHORT)
        [Browsable(false)]
        public UIElement ContentLine { get { return (UIElement)GetValue(ContentLineProperty); } set { SetValue(ContentLineProperty, value); } }
        public static readonly DependencyProperty ContentLineProperty = DependencyProperty.Register("ContentLine", typeof(UIElement), typeof(BasePaletteAnimation), new PropertyMetadata(null));
        #endregion

        #region AdjustTop (DP SHORT)
        [Category("Palette")]
        public double AdjustTop { get { return (double)GetValue(AdjustTopProperty); } set { SetValue(AdjustTopProperty, value); } }
        public static readonly DependencyProperty AdjustTopProperty =
            DependencyProperty.Register("AdjustTop",
                typeof(double), typeof(BasePaletteAnimation),
                new PropertyMetadata(0D));
        #endregion

        #region CornerRadius (DP SHORT)
        [Category("Palette")]
        public CornerRadius CornerRadius { get { return (CornerRadius)GetValue(CornerRadiusProperty); } set { SetValue(CornerRadiusProperty, value); } }
        public static readonly DependencyProperty CornerRadiusProperty =
            DependencyProperty.Register("CornerRadius",
                typeof(CornerRadius), typeof(BasePaletteAnimation),
                new PropertyMetadata(new CornerRadius(2.0)));
        #endregion

        #region AnimationTime (DP SHORT)
        [Category("Palette")]
        public double AnimationTime { get { return (double)GetValue(AnimationTimeProperty); } set { SetValue(AnimationTimeProperty, value); } }
        public static readonly DependencyProperty AnimationTimeProperty = DependencyProperty.Register("AnimationTime", typeof(double), typeof(BasePaletteAnimation), new PropertyMetadata(300D));
        #endregion

        #endregion

        #region Animation
        Storyboard storyboard = null;
        protected void StartAnimation(UIElement old)
        {
            SizeChanged();

            if (paletteBack == null) return;

            #region Stop current animation if necessary
            if (storyboard != null)
            {
                storyboard.Stop();
                DoEndAnim(old);
            }
            #endregion

            #region Start Animation
            paletteBack.Content = GetContent;
            storyboard = new Storyboard();
            // frontTop 1 to 0
            {
                DoubleAnimation da = new DoubleAnimation();
                Storyboard.SetTargetProperty(da, new PropertyPath("RenderTransform.ScaleY"));
                Storyboard.SetTarget(da, frontTop);
                da.Duration = new Duration(TimeSpan.FromMilliseconds(AnimationTime));
                da.BeginTime = TimeSpan.FromSeconds(0);
                da.From = 1;
                da.To = 0;
                storyboard.Children.Add(da);
            }
            // frontbottom visibility = hidden
            {
                ObjectAnimationUsingKeyFrames animate = new ObjectAnimationUsingKeyFrames();
                animate.BeginTime = TimeSpan.FromMilliseconds(AnimationTime);
                animate.Duration = new TimeSpan(0);
                Storyboard.SetTargetProperty(animate, new PropertyPath("Visibility"));
                Storyboard.SetTarget(animate, frontBottom);
                DiscreteObjectKeyFrame kf1 = new DiscreteObjectKeyFrame(
                    Visibility.Hidden,
                    new TimeSpan(0, 0, 0, 0, 0));
                animate.KeyFrames.Add(kf1);
                storyboard.Children.Add(animate);
            }
            #endregion
            storyboard.Completed += delegate { DoEndAnim(); };
            storyboard.Begin();
        }

        protected void DoEndAnim(UIElement element = null)
        {
            UIElement ui = element ?? GetContent;
            paletteTop.Content = ui.CloneElement();
            paletteBottom.Content = ui.CloneElement();
            // 
            Storyboard sb = new Storyboard();
            // frontTop to 1 (immediat)
            {
                DoubleAnimation da = new DoubleAnimation();
                Storyboard.SetTargetProperty(da, new PropertyPath("RenderTransform.ScaleY"));
                Storyboard.SetTarget(da, frontTop);
                da.Duration = new Duration(TimeSpan.FromMilliseconds(0));
                da.BeginTime = TimeSpan.FromSeconds(0);
                da.From = 0;
                da.To = 1;
                sb.Children.Add(da);
            }
            // frontbottom visibility (immediat)
            {
                ObjectAnimationUsingKeyFrames animate = new ObjectAnimationUsingKeyFrames();
                animate.BeginTime = TimeSpan.FromMilliseconds(AnimationTime);
                animate.Duration = new TimeSpan(0);
                Storyboard.SetTargetProperty(animate, new PropertyPath("Visibility"));
                Storyboard.SetTarget(animate, frontBottom);
                DiscreteObjectKeyFrame kf1 = new DiscreteObjectKeyFrame(
                    Visibility.Visible,
                    new TimeSpan(0, 0, 0, 0, 0));
                animate.KeyFrames.Add(kf1);
                sb.Children.Add(animate);
            }
            // start
            sb.Completed += delegate
            {
                sb = null;
                if (element == null) RaiseEvents();
            };
            sb.Begin();
        }
        #endregion

        protected void RaiseEvents()
        {
            OnEndAnim();
            OnEndAnimation();
        }
    }
}
